@@ -1,47 +1,12 @@
-/****************************************************************************
-**
-** Copyright (C) 2016 The Qt Company Ltd.
-** Contact: https://www.qt.io/licensing/
-**
-** This file is part of the QtQml module of the Qt Toolkit.
-**
-** $QT_BEGIN_LICENSE:LGPL$
-** Commercial License Usage
-** Licensees holding valid commercial Qt licenses may use this file in
-** accordance with the commercial license agreement provided with the
-** Software or, alternatively, in accordance with the terms contained in
-** a written agreement between you and The Qt Company. For licensing terms
-** and conditions see https://www.qt.io/terms-conditions. For further
-** information use the contact form at https://www.qt.io/contact-us.
-**
-** GNU Lesser General Public License Usage
-** Alternatively, this file may be used under the terms of the GNU Lesser
-** General Public License version 3 as published by the Free Software
-** Foundation and appearing in the file LICENSE.LGPL3 included in the
-** packaging of this file. Please review the following information to
-** ensure the GNU Lesser General Public License version 3 requirements
-** will be met: https://www.gnu.org/licenses/lgpl-3.0.html.
-**
-** GNU General Public License Usage
-** Alternatively, this file may be used under the terms of the GNU
-** General Public License version 2.0 or (at your option) the GNU General
-** Public license version 3 or any later version approved by the KDE Free
-** Qt Foundation. The licenses are as published by the Free Software
-** Foundation and appearing in the file LICENSE.GPL2 and LICENSE.GPL3
-** included in the packaging of this file. Please review the following
-** information to ensure the GNU General Public License requirements will
-** be met: https://www.gnu.org/licenses/gpl-2.0.html and
-** https://www.gnu.org/licenses/gpl-3.0.html.
-**
-** $QT_END_LICENSE$
-**
-****************************************************************************/
+// Copyright (C) 2021 The Qt Company Ltd.
+// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR LGPL-3.0-only OR GPL-2.0-only OR GPL-3.0-only
 
 #ifndef QQML_H
 #define QQML_H
 
 #include <QtQml/qqmlprivate.h>
 #include <QtQml/qjsvalue.h>
+#include <QtQml/qqmlregistration.h>
 
 #include <QtCore/qbytearray.h>
 #include <QtCore/qmetaobject.h>
@@ -51,14 +16,8 @@
 #define QML_VERSION     0x020000
 #define QML_VERSION_STR "2.0"
 
-#define QML_PRIVATE_NAMESPACE \
-    QT_PREPEND_NAMESPACE(QQmlPrivate)
-
-#define QML_REGISTER_TYPES_AND_REVISIONS \
-    QT_PREPEND_NAMESPACE(qmlRegisterTypesAndRevisions)
-
 #define QML_DECLARE_TYPE(TYPE) \
-    Q_DECLARE_METATYPE(TYPE *) \
+    Q_DECLARE_METATYPE(TYPE*) \
     Q_DECLARE_METATYPE(QQmlListProperty<TYPE>)
 
 #define QML_DECLARE_TYPE_HASMETATYPE(TYPE) \
@@ -70,94 +29,6 @@
 #define QML_DECLARE_INTERFACE_HASMETATYPE(INTERFACE) \
     QML_DECLARE_TYPE_HASMETATYPE(INTERFACE)
 
-#define QML_ELEMENT \
-    Q_CLASSINFO("QML.Element", "auto")
-
-#define QML_ANONYMOUS \
-    Q_CLASSINFO("QML.Element", "anonymous")
-
-#define QML_NAMED_ELEMENT(NAME) \
-    Q_CLASSINFO("QML.Element", #NAME)
-
-#define QML_VALUE_TYPE(NAME) \
-    Q_CLASSINFO("QML.Element", #NAME) \
-    QML_UNCREATABLE("Value types cannot be created.")
-
-#define QML_UNCREATABLE(REASON) \
-    Q_CLASSINFO("QML.Creatable", "false") \
-    Q_CLASSINFO("QML.UncreatableReason", REASON)
-
-#define QML_SINGLETON \
-    Q_CLASSINFO("QML.Singleton", "true") \
-    enum class QmlIsSingleton {yes = true}; \
-    template<typename, typename> friend struct QML_PRIVATE_NAMESPACE::QmlSingleton; \
-    template<typename T, typename... Args> \
-    friend void QML_REGISTER_TYPES_AND_REVISIONS(const char *uri, int versionMajor, QList<int> *);
-
-#define QML_SEQUENTIAL_CONTAINER(VALUE_TYPE) \
-    Q_CLASSINFO("QML.Sequence", #VALUE_TYPE) \
-    using QmlSequenceValueType = VALUE_TYPE; \
-    enum class QmlIsSequence {yes = true}; \
-    template<typename, typename> friend struct QML_PRIVATE_NAMESPACE::QmlSequence; \
-    template<typename T, typename... Args> \
-    friend void QML_REGISTER_TYPES_AND_REVISIONS(const char *uri, int versionMajor, QList<int> *);
-
-#define QML_ADDED_IN_MINOR_VERSION(VERSION) \
-    Q_CLASSINFO("QML.AddedInVersion", Q_REVISION(VERSION))
-
-#define QML_ADDED_IN_VERSION(MAJOR, MINOR) \
-    Q_CLASSINFO("QML.AddedInVersion", Q_REVISION(MAJOR, MINOR))
-
-#define QML_REMOVED_IN_MINOR_VERSION(VERSION) \
-    Q_CLASSINFO("QML.RemovedInVersion", Q_REVISION(VERSION))
-
-#define QML_REMOVED_IN_VERSION(MAJOR, MINOR) \
-    Q_CLASSINFO("QML.RemovedInVersion", Q_REVISION(MAJOR, MINOR))
-
-#define QML_ATTACHED(ATTACHED_TYPE) \
-    Q_CLASSINFO("QML.Attached", #ATTACHED_TYPE) \
-    using QmlAttachedType = ATTACHED_TYPE; \
-    template<class, class, bool> friend struct QML_PRIVATE_NAMESPACE::QmlAttached; \
-    template<class> friend struct QML_PRIVATE_NAMESPACE::QmlAttachedAccessor;
-
-#define QML_EXTENDED(EXTENDED_TYPE) \
-    Q_CLASSINFO("QML.Extended", #EXTENDED_TYPE) \
-    using QmlExtendedType = EXTENDED_TYPE; \
-    template<class, class> friend struct QML_PRIVATE_NAMESPACE::QmlExtended; \
-    template<typename T, typename... Args> \
-    friend void QML_REGISTER_TYPES_AND_REVISIONS(const char *uri, int versionMajor, QList<int> *);
-
-#define QML_EXTENDED_NAMESPACE(EXTENDED_NAMESPACE) \
-    Q_CLASSINFO("QML.Extended", #EXTENDED_NAMESPACE) \
-    static constexpr const QMetaObject *qmlExtendedNamespace() { return &EXTENDED_NAMESPACE::staticMetaObject; } \
-    template<class, class> friend struct QML_PRIVATE_NAMESPACE::QmlExtendedNamespace; \
-    template<typename T, typename... Args> \
-    friend void QML_REGISTER_TYPES_AND_REVISIONS(const char *uri, int versionMajor, QList<int> *);
-
-#define QML_FOREIGN(FOREIGN_TYPE) \
-    Q_CLASSINFO("QML.Foreign", #FOREIGN_TYPE) \
-    using QmlForeignType = FOREIGN_TYPE; \
-    template<class, class> friend struct QML_PRIVATE_NAMESPACE::QmlResolved; \
-    template<typename T, typename... Args> \
-    friend void QML_REGISTER_TYPES_AND_REVISIONS(const char *uri, int versionMajor, QList<int> *);
-
-#define QML_FOREIGN_NAMESPACE(FOREIGN_NAMESPACE) \
-    Q_CLASSINFO("QML.Foreign", #FOREIGN_NAMESPACE)
-
-#define QML_INTERFACE \
-    Q_CLASSINFO("QML.Element", "anonymous") \
-    enum class QmlIsInterface {yes = true}; \
-    template<typename, typename> friend struct QML_PRIVATE_NAMESPACE::QmlInterface; \
-    template<typename T, typename... Args> \
-    friend void QML_REGISTER_TYPES_AND_REVISIONS(const char *uri, int versionMajor, QList<int> *);
-
-#define QML_IMPLEMENTS_INTERFACES(INTERFACES) \
-    Q_INTERFACES(INTERFACES) \
-    enum class QmlIsInterface {yes = false};
-
-#define QML_UNAVAILABLE \
-    QML_FOREIGN(QQmlTypeNotAvailable)
-
 enum { /* TYPEINFO flags */
     QML_HAS_ATTACHED_PROPERTIES = 0x01
 };
@@ -185,7 +56,7 @@ template<typename T>
 int qmlRegisterAnonymousType(const char *uri, int versionMajor)
 {
     QQmlPrivate::RegisterType type = {
-        0,
+        QQmlPrivate::RegisterType::CurrentVersion,
         QQmlPrivate::QmlMetaType<T>::self(),
         QQmlPrivate::QmlMetaType<T>::list(),
         0,
@@ -206,20 +77,78 @@ int qmlRegisterAnonymousType(const char *uri, int versionMajor)
         nullptr, nullptr,
 
         nullptr,
-        QTypeRevision::zero()
+        QTypeRevision::zero(),
+        QQmlPrivate::StaticCastSelector<T,QQmlFinalizerHook>::cast(),
+        QQmlPrivate::ValueTypeCreationMethod::None,
     };
 
     return QQmlPrivate::qmlregister(QQmlPrivate::TypeRegistration, &type);
 }
 
+//! \internal
+template<typename T, int metaObjectRevisionMinor>
+int qmlRegisterAnonymousType(const char *uri, int versionMajor)
+{
+    QQmlPrivate::RegisterType type = {
+        QQmlPrivate::RegisterType::CurrentVersion,
+        QQmlPrivate::QmlMetaType<T>::self(),
+        QQmlPrivate::QmlMetaType<T>::list(),
+        0,
+        nullptr,
+        nullptr,
+        QString(),
+        QQmlPrivate::ValueType<T, void>::create,
+
+        uri,
+        QTypeRevision::fromVersion(versionMajor, 0),
+        nullptr,
+        QQmlPrivate::StaticMetaObject<T>::staticMetaObject(),
+
+        QQmlPrivate::attachedPropertiesFunc<T>(),
+        QQmlPrivate::attachedPropertiesMetaObject<T>(),
+
+        QQmlPrivate::StaticCastSelector<T, QQmlParserStatus>::cast(),
+        QQmlPrivate::StaticCastSelector<T, QQmlPropertyValueSource>::cast(),
+        QQmlPrivate::StaticCastSelector<T, QQmlPropertyValueInterceptor>::cast(),
+
+        nullptr,
+        nullptr,
+
+        nullptr,
+        QTypeRevision::fromMinorVersion(metaObjectRevisionMinor),
+        QQmlPrivate::StaticCastSelector<T, QQmlFinalizerHook>::cast(),
+        QQmlPrivate::ValueTypeCreationMethod::None,
+    };
+
+    return QQmlPrivate::qmlregister(QQmlPrivate::TypeRegistration, &type);
+}
+
+//! \internal
+template<typename T>
+void qmlRegisterAnonymousTypesAndRevisions(const char *uri, int versionMajor)
+{
+    // Anonymous types are not creatable, no need to warn about missing acceptable constructors.
+    QQmlPrivate::qmlRegisterTypeAndRevisions<T, void>(
+            uri, versionMajor, QQmlPrivate::StaticMetaObject<T>::staticMetaObject(), nullptr,
+            nullptr, true);
+}
+
+class QQmlTypeNotAvailable : public QObject
+{
+    Q_OBJECT
+    QML_NAMED_ELEMENT(TypeNotAvailable)
+    QML_ADDED_IN_VERSION(2, 15)
+    QML_UNCREATABLE("Type not available.")
+};
+
 int Q_QML_EXPORT qmlRegisterTypeNotAvailable(const char *uri, int versionMajor, int versionMinor,
-                                             const char *qmlName, const QString& message);
+                                             const char *qmlName, const QString &message);
 
 template<typename T>
 int qmlRegisterUncreatableType(const char *uri, int versionMajor, int versionMinor, const char *qmlName, const QString& reason)
 {
     QQmlPrivate::RegisterType type = {
-        0,
+        QQmlPrivate::RegisterType::CurrentVersion,
         QQmlPrivate::QmlMetaType<T>::self(),
         QQmlPrivate::QmlMetaType<T>::list(),
         0,
@@ -241,7 +170,9 @@ int qmlRegisterUncreatableType(const char *uri, int versionMajor, int versionMin
         nullptr, nullptr,
 
         nullptr,
-        QTypeRevision::zero()
+        QTypeRevision::zero(),
+        QQmlPrivate::StaticCastSelector<T,QQmlFinalizerHook>::cast(),
+        QQmlPrivate::ValueTypeCreationMethod::None,
     };
 
     return QQmlPrivate::qmlregister(QQmlPrivate::TypeRegistration, &type);
@@ -251,7 +182,7 @@ template<typename T, int metaObjectRevision>
 int qmlRegisterUncreatableType(const char *uri, int versionMajor, int versionMinor, const char *qmlName, const QString& reason)
 {
     QQmlPrivate::RegisterType type = {
-        0,
+        QQmlPrivate::RegisterType::CurrentVersion,
         QQmlPrivate::QmlMetaType<T>::self(),
         QQmlPrivate::QmlMetaType<T>::list(),
         0,
@@ -273,7 +204,9 @@ int qmlRegisterUncreatableType(const char *uri, int versionMajor, int versionMin
         nullptr, nullptr,
 
         nullptr,
-        QTypeRevision::fromMinorVersion(metaObjectRevision)
+        QTypeRevision::fromMinorVersion(metaObjectRevision),
+        QQmlPrivate::StaticCastSelector<T,QQmlFinalizerHook>::cast(),
+        QQmlPrivate::ValueTypeCreationMethod::None,
     };
 
     return QQmlPrivate::qmlregister(QQmlPrivate::TypeRegistration, &type);
@@ -290,7 +223,7 @@ int qmlRegisterExtendedUncreatableType(const char *uri, int versionMajor, int ve
     }
 
     QQmlPrivate::RegisterType type = {
-        0,
+        QQmlPrivate::RegisterType::CurrentVersion,
         QQmlPrivate::QmlMetaType<T>::self(),
         QQmlPrivate::QmlMetaType<T>::list(),
         0,
@@ -312,7 +245,9 @@ int qmlRegisterExtendedUncreatableType(const char *uri, int versionMajor, int ve
         QQmlPrivate::ExtendedType<E>::createParent, QQmlPrivate::ExtendedType<E>::staticMetaObject(),
 
         nullptr,
-        QTypeRevision::zero()
+        QTypeRevision::zero(),
+        QQmlPrivate::StaticCastSelector<T,QQmlFinalizerHook>::cast(),
+        QQmlPrivate::ValueTypeCreationMethod::None,
     };
 
     return QQmlPrivate::qmlregister(QQmlPrivate::TypeRegistration, &type);
@@ -329,7 +264,7 @@ int qmlRegisterExtendedUncreatableType(const char *uri, int versionMajor, int ve
     }
 
     QQmlPrivate::RegisterType type = {
-        0,
+        QQmlPrivate::RegisterType::CurrentVersion,
         QQmlPrivate::QmlMetaType<T>::self(),
         QQmlPrivate::QmlMetaType<T>::list(),
         0,
@@ -351,7 +286,9 @@ int qmlRegisterExtendedUncreatableType(const char *uri, int versionMajor, int ve
         QQmlPrivate::ExtendedType<E>::createParent, QQmlPrivate::ExtendedType<E>::staticMetaObject(),
 
         nullptr,
-        QTypeRevision::fromMinorVersion(metaObjectRevision)
+        QTypeRevision::fromMinorVersion(metaObjectRevision),
+        QQmlPrivate::StaticCastSelector<T,QQmlFinalizerHook>::cast(),
+        QQmlPrivate::ValueTypeCreationMethod::None,
     };
 
     return QQmlPrivate::qmlregister(QQmlPrivate::TypeRegistration, &type);
@@ -362,8 +299,12 @@ Q_QML_EXPORT int qmlRegisterUncreatableMetaObject(const QMetaObject &staticMetaO
 template<typename T>
 int qmlRegisterType(const char *uri, int versionMajor, int versionMinor, const char *qmlName)
 {
+    static_assert(!std::is_abstract_v<T>,
+        "It is not possible to register an abstract type with qmlRegisterType. "
+        "Maybe you wanted qmlRegisterUncreatableType or qmlRegisterInterface?");
+
     QQmlPrivate::RegisterType type = {
-        0,
+        QQmlPrivate::RegisterType::CurrentVersion,
         QQmlPrivate::QmlMetaType<T>::self(),
         QQmlPrivate::QmlMetaType<T>::list(),
         sizeof(T), QQmlPrivate::Constructors<T>::createInto, nullptr,
@@ -383,7 +324,9 @@ int qmlRegisterType(const char *uri, int versionMajor, int versionMinor, const c
         nullptr, nullptr,
 
         nullptr,
-        QTypeRevision::zero()
+        QTypeRevision::zero(),
+        QQmlPrivate::StaticCastSelector<T,QQmlFinalizerHook>::cast(),
+        QQmlPrivate::ValueTypeCreationMethod::None,
     };
 
     return QQmlPrivate::qmlregister(QQmlPrivate::TypeRegistration, &type);
@@ -392,8 +335,12 @@ int qmlRegisterType(const char *uri, int versionMajor, int versionMinor, const c
 template<typename T, int metaObjectRevision>
 int qmlRegisterType(const char *uri, int versionMajor, int versionMinor, const char *qmlName)
 {
+    static_assert(!std::is_abstract_v<T>,
+        "It is not possible to register an abstract type with qmlRegisterType. "
+        "Maybe you wanted qmlRegisterUncreatableType or qmlRegisterInterface?");
+
     QQmlPrivate::RegisterType type = {
-        0,
+        QQmlPrivate::RegisterType::CurrentVersion,
         QQmlPrivate::QmlMetaType<T>::self(),
         QQmlPrivate::QmlMetaType<T>::list(),
         sizeof(T), QQmlPrivate::Constructors<T>::createInto, nullptr,
@@ -413,7 +360,9 @@ int qmlRegisterType(const char *uri, int versionMajor, int versionMinor, const c
         nullptr, nullptr,
 
         nullptr,
-        QTypeRevision::fromMinorVersion(metaObjectRevision)
+        QTypeRevision::fromMinorVersion(metaObjectRevision),
+        QQmlPrivate::StaticCastSelector<T,QQmlFinalizerHook>::cast(),
+        QQmlPrivate::ValueTypeCreationMethod::None,
     };
 
     return QQmlPrivate::qmlregister(QQmlPrivate::TypeRegistration, &type);
@@ -423,7 +372,7 @@ template<typename T, int metaObjectRevision>
 int qmlRegisterRevision(const char *uri, int versionMajor, int versionMinor)
 {
     QQmlPrivate::RegisterType type = {
-        0,
+        QQmlPrivate::RegisterType::CurrentVersion,
         QQmlPrivate::QmlMetaType<T>::self(),
         QQmlPrivate::QmlMetaType<T>::list(),
         sizeof(T), QQmlPrivate::Constructors<T>::createInto, nullptr,
@@ -443,7 +392,9 @@ int qmlRegisterRevision(const char *uri, int versionMajor, int versionMinor)
         nullptr, nullptr,
 
         nullptr,
-        QTypeRevision::fromMinorVersion(metaObjectRevision)
+        QTypeRevision::fromMinorVersion(metaObjectRevision),
+        QQmlPrivate::StaticCastSelector<T,QQmlFinalizerHook>::cast(),
+        QQmlPrivate::ValueTypeCreationMethod::None,
     };
 
     return QQmlPrivate::qmlregister(QQmlPrivate::TypeRegistration, &type);
@@ -452,8 +403,15 @@ int qmlRegisterRevision(const char *uri, int versionMajor, int versionMinor)
 template<typename T, typename E>
 int qmlRegisterExtendedType(const char *uri, int versionMajor)
 {
+    static_assert(!std::is_abstract_v<T>,
+        "It is not possible to register an extension to an abstract type with qmlRegisterExtendedType.");
+
+    static_assert(!std::is_abstract_v<E>,
+        "It is not possible to register an abstract type with qmlRegisterExtendedType. "
+        "Maybe you wanted qmlRegisterExtendedUncreatableType?");
+
     QQmlPrivate::RegisterType type = {
-        0,
+        QQmlPrivate::RegisterType::CurrentVersion,
         QQmlPrivate::QmlMetaType<T>::self(),
         QQmlPrivate::QmlMetaType<T>::list(),
         0,
@@ -475,7 +433,9 @@ int qmlRegisterExtendedType(const char *uri, int versionMajor)
         QQmlPrivate::ExtendedType<E>::createParent, QQmlPrivate::ExtendedType<E>::staticMetaObject(),
 
         nullptr,
-        QTypeRevision::zero()
+        QTypeRevision::zero(),
+        QQmlPrivate::StaticCastSelector<T,QQmlFinalizerHook>::cast(),
+        QQmlPrivate::ValueTypeCreationMethod::None,
     };
 
     return QQmlPrivate::qmlregister(QQmlPrivate::TypeRegistration, &type);
@@ -485,6 +445,13 @@ template<typename T, typename E>
 int qmlRegisterExtendedType(const char *uri, int versionMajor, int versionMinor,
                             const char *qmlName)
 {
+    static_assert(!std::is_abstract_v<T>,
+        "It is not possible to register an extension to an abstract type with qmlRegisterExtendedType.");
+
+    static_assert(!std::is_abstract_v<E>,
+        "It is not possible to register an abstract type with qmlRegisterExtendedType. "
+        "Maybe you wanted qmlRegisterExtendedUncreatableType?");
+
     QQmlAttachedPropertiesFunc attached = QQmlPrivate::attachedPropertiesFunc<E>();
     const QMetaObject * attachedMetaObject = QQmlPrivate::attachedPropertiesMetaObject<E>();
     if (!attached) {
@@ -493,7 +460,7 @@ int qmlRegisterExtendedType(const char *uri, int versionMajor, int versionMinor,
     }
 
     QQmlPrivate::RegisterType type = {
-        0,
+        QQmlPrivate::RegisterType::CurrentVersion,
         QQmlPrivate::QmlMetaType<T>::self(),
         QQmlPrivate::QmlMetaType<T>::list(),
         sizeof(T), QQmlPrivate::Constructors<T>::createInto, nullptr,
@@ -513,7 +480,9 @@ int qmlRegisterExtendedType(const char *uri, int versionMajor, int versionMinor,
         QQmlPrivate::ExtendedType<E>::createParent, QQmlPrivate::ExtendedType<E>::staticMetaObject(),
 
         nullptr,
-        QTypeRevision::zero()
+        QTypeRevision::zero(),
+        QQmlPrivate::StaticCastSelector<T,QQmlFinalizerHook>::cast(),
+        QQmlPrivate::ValueTypeCreationMethod::None,
     };
 
     return QQmlPrivate::qmlregister(QQmlPrivate::TypeRegistration, &type);
@@ -541,8 +510,12 @@ template<typename T>
 int qmlRegisterCustomType(const char *uri, int versionMajor, int versionMinor,
                           const char *qmlName, QQmlCustomParser *parser)
 {
+    static_assert(!std::is_abstract_v<T>,
+        "It is not possible to register an abstract type with qmlRegisterCustomType. "
+        "Maybe you wanted qmlRegisterUncreatableType or qmlRegisterInterface?");
+
     QQmlPrivate::RegisterType type = {
-        0,
+        QQmlPrivate::RegisterType::CurrentVersion,
         QQmlPrivate::QmlMetaType<T>::self(),
         QQmlPrivate::QmlMetaType<T>::list(),
         sizeof(T), QQmlPrivate::Constructors<T>::createInto, nullptr,
@@ -562,7 +535,9 @@ int qmlRegisterCustomType(const char *uri, int versionMajor, int versionMinor,
         nullptr, nullptr,
 
         parser,
-        QTypeRevision::zero()
+        QTypeRevision::zero(),
+        QQmlPrivate::StaticCastSelector<T,QQmlFinalizerHook>::cast(),
+        QQmlPrivate::ValueTypeCreationMethod::None,
     };
 
     return QQmlPrivate::qmlregister(QQmlPrivate::TypeRegistration, &type);
@@ -572,8 +547,12 @@ template<typename T, int metaObjectRevision>
 int qmlRegisterCustomType(const char *uri, int versionMajor, int versionMinor,
                           const char *qmlName, QQmlCustomParser *parser)
 {
+    static_assert(!std::is_abstract_v<T>,
+        "It is not possible to register an abstract type with qmlRegisterCustomType. "
+        "Maybe you wanted qmlRegisterUncreatableType or qmlRegisterInterface?");
+
     QQmlPrivate::RegisterType type = {
-        0,
+        QQmlPrivate::RegisterType::CurrentVersion,
         QQmlPrivate::QmlMetaType<T>::self(),
         QQmlPrivate::QmlMetaType<T>::list(),
         sizeof(T), QQmlPrivate::Constructors<T>::createInto, nullptr,
@@ -593,7 +572,9 @@ int qmlRegisterCustomType(const char *uri, int versionMajor, int versionMinor,
         nullptr, nullptr,
 
         parser,
-        QTypeRevision::fromMinorVersion(metaObjectRevision)
+        QTypeRevision::fromMinorVersion(metaObjectRevision),
+        QQmlPrivate::StaticCastSelector<T,QQmlFinalizerHook>::cast(),
+        QQmlPrivate::ValueTypeCreationMethod::None,
     };
 
     return QQmlPrivate::qmlregister(QQmlPrivate::TypeRegistration, &type);
@@ -603,6 +584,12 @@ template<typename T, typename E>
 int qmlRegisterCustomExtendedType(const char *uri, int versionMajor, int versionMinor,
                           const char *qmlName, QQmlCustomParser *parser)
 {
+    static_assert(!std::is_abstract_v<T>,
+        "It is not possible to register an extension to an abstract type with qmlRegisterCustomExtendedType.");
+
+    static_assert(!std::is_abstract_v<E>,
+        "It is not possible to register an abstract type with qmlRegisterCustomExtendedType.");
+
     QQmlAttachedPropertiesFunc attached = QQmlPrivate::attachedPropertiesFunc<E>();
     const QMetaObject * attachedMetaObject = QQmlPrivate::attachedPropertiesMetaObject<E>();
     if (!attached) {
@@ -611,7 +598,7 @@ int qmlRegisterCustomExtendedType(const char *uri, int versionMajor, int version
     }
 
     QQmlPrivate::RegisterType type = {
-        0,
+        QQmlPrivate::RegisterType::CurrentVersion,
         QQmlPrivate::QmlMetaType<T>::self(),
         QQmlPrivate::QmlMetaType<T>::list(),
         sizeof(T), QQmlPrivate::Constructors<T>::createInto, nullptr,
@@ -631,7 +618,9 @@ int qmlRegisterCustomExtendedType(const char *uri, int versionMajor, int version
         QQmlPrivate::ExtendedType<E>::createParent, QQmlPrivate::ExtendedType<E>::staticMetaObject(),
 
         parser,
-        QTypeRevision::zero()
+        QTypeRevision::zero(),
+        QQmlPrivate::StaticCastSelector<T,QQmlFinalizerHook>::cast(),
+        QQmlPrivate::ValueTypeCreationMethod::None,
     };
 
     return QQmlPrivate::qmlregister(QQmlPrivate::TypeRegistration, &type);
@@ -649,6 +638,7 @@ Q_QML_EXPORT QQmlAttachedPropertiesFunc qmlAttachedPropertiesFunction(QObject *,
                                                                       const QMetaObject *);
 Q_QML_EXPORT QObject *qmlAttachedPropertiesObject(QObject *, QQmlAttachedPropertiesFunc func,
                                                   bool create = true);
+Q_QML_EXPORT QObject *qmlExtendedObject(QObject *);
 
 //The C++ version of protected namespaces in qmldir
 Q_QML_EXPORT bool qmlProtectModule(const char* uri, int majVersion);
@@ -752,7 +742,7 @@ inline auto qmlRegisterSingletonInstance(const char *uri, int versionMajor, int
                                          const char *typeName, T *cppObject) -> typename std::enable_if<std::is_base_of<QObject, T>::value, int>::type
 #endif
 {
-    QQmlPrivate::SingletonFunctor registrationFunctor;
+    QQmlPrivate::SingletonInstanceFunctor registrationFunctor;
     registrationFunctor.m_object = cppObject;
     return qmlRegisterSingletonType<T>(uri, versionMajor, versionMinor, typeName, registrationFunctor);
 }
@@ -798,6 +788,9 @@ inline int qmlRegisterType(const QUrl &url, const char *uri, int versionMajor, i
 template<typename Container>
 inline int qmlRegisterAnonymousSequentialContainer(const char *uri, int versionMajor)
 {
+    static_assert(!std::is_abstract_v<Container>,
+        "It is not possible to register an abstract container with qmlRegisterAnonymousSequentialContainer.");
+
     QQmlPrivate::RegisterSequentialContainer type = {
         0,
         uri,
@@ -811,14 +804,56 @@ inline int qmlRegisterAnonymousSequentialContainer(const char *uri, int versionM
     return QQmlPrivate::qmlregister(QQmlPrivate::SequentialContainerRegistration, &type);
 }
 
-template<class T, class Resolved, class Extended, bool Singleton, bool Interface, bool Sequence>
+template<class T, class Resolved, class Extended, bool Singleton, bool Interface, bool Sequence, bool Uncreatable>
 struct QmlTypeAndRevisionsRegistration;
 
 template<class T, class Resolved, class Extended>
-struct QmlTypeAndRevisionsRegistration<T, Resolved, Extended, false, false, false> {
+struct QmlTypeAndRevisionsRegistration<T, Resolved, Extended, false, false, false, false> {
     static void registerTypeAndRevisions(const char *uri, int versionMajor, QList<int> *qmlTypeIds,
                                          const QMetaObject *extension)
     {
+#if QT_DEPRECATED_SINCE(6, 4)
+        // ### Qt7: Remove the warnings, and leave only the static asserts below.
+        if constexpr (!QQmlPrivate::QmlMetaType<Resolved>::hasAcceptableCtors()) {
+            QQmlPrivate::qmlRegistrationWarning(QQmlPrivate::UnconstructibleType,
+                                                QMetaType::fromType<Resolved>());
+        }
+
+        if constexpr (!std::is_base_of_v<QObject, Resolved>
+                && QQmlTypeInfo<T>::hasAttachedProperties) {
+            QQmlPrivate::qmlRegistrationWarning(QQmlPrivate::NonQObjectWithAtached,
+                                                QMetaType::fromType<Resolved>());
+        }
+#else
+        static_assert(QQmlPrivate::QmlMetaType<Resolved>::hasAcceptableCtors(),
+                      "This type is neither a QObject, nor default- and copy-constructible, nor"
+                      "uncreatable.\n"
+                      "You should not use it as a QML type.");
+        static_assert(std::is_base_of_v<QObject, Resolved>
+                || !QQmlTypeInfo<Resolved>::hasAttachedProperties);
+#endif
+        QQmlPrivate::qmlRegisterTypeAndRevisions<Resolved, Extended>(
+                    uri, versionMajor, QQmlPrivate::StaticMetaObject<T>::staticMetaObject(),
+                    qmlTypeIds, extension);
+    }
+};
+
+template<class T, class Resolved, class Extended>
+struct QmlTypeAndRevisionsRegistration<T, Resolved, Extended, false, false, false, true> {
+    static void registerTypeAndRevisions(const char *uri, int versionMajor, QList<int> *qmlTypeIds,
+                                         const QMetaObject *extension)
+    {
+#if QT_DEPRECATED_SINCE(6, 4)
+        // ### Qt7: Remove the warning, and leave only the static assert below.
+        if constexpr (!std::is_base_of_v<QObject, Resolved>
+                && QQmlTypeInfo<Resolved>::hasAttachedProperties) {
+            QQmlPrivate::qmlRegistrationWarning(QQmlPrivate::NonQObjectWithAtached,
+                                                QMetaType::fromType<Resolved>());
+        }
+#else
+        static_assert(std::is_base_of_v<QObject, Resolved>
+                || !QQmlTypeInfo<Resolved>::hasAttachedProperties);
+#endif
         QQmlPrivate::qmlRegisterTypeAndRevisions<Resolved, Extended>(
                     uri, versionMajor, QQmlPrivate::StaticMetaObject<T>::staticMetaObject(),
                     qmlTypeIds, extension);
@@ -826,10 +861,11 @@ struct QmlTypeAndRevisionsRegistration<T, Resolved, Extended, false, false, fals
 };
 
 template<class T, class Resolved>
-struct QmlTypeAndRevisionsRegistration<T, Resolved, void, false, false, true> {
+struct QmlTypeAndRevisionsRegistration<T, Resolved, void, false, false, true, true> {
     static void registerTypeAndRevisions(const char *uri, int versionMajor, QList<int> *qmlTypeIds,
                                          const QMetaObject *)
     {
+        // Sequences have to be anonymous for now, which implies uncreatable.
         QQmlPrivate::qmlRegisterSequenceAndRevisions<Resolved>(
                     uri, versionMajor, QQmlPrivate::StaticMetaObject<T>::staticMetaObject(),
                     qmlTypeIds);
@@ -837,18 +873,45 @@ struct QmlTypeAndRevisionsRegistration<T, Resolved, void, false, false, true> {
 };
 
 template<class T, class Resolved, class Extended>
-struct QmlTypeAndRevisionsRegistration<T, Resolved, Extended, true, false, false> {
+struct QmlTypeAndRevisionsRegistration<T, Resolved, Extended, true, false, false, false> {
     static void registerTypeAndRevisions(const char *uri, int versionMajor, QList<int> *qmlTypeIds,
                                          const QMetaObject *extension)
     {
-        QQmlPrivate::qmlRegisterSingletonAndRevisions<Resolved, Extended>(
+#if QT_DEPRECATED_SINCE(6, 4)
+        // ### Qt7: Remove the warning, and leave only the static assert below.
+        if constexpr (QQmlPrivate::singletonConstructionMode<Resolved, T>()
+                == QQmlPrivate::SingletonConstructionMode::None) {
+            QQmlPrivate::qmlRegistrationWarning(QQmlPrivate::UnconstructibleSingleton,
+                                                QMetaType::fromType<Resolved>());
+        }
+#else
+        static_assert(QQmlPrivate::singletonConstructionMode<Resolved, T>()
+                        != QQmlPrivate::SingletonConstructionMode::None,
+                      "A singleton needs either a default constructor or, when adding a default "
+                      "constructor is infeasible, a public static "
+                      "create(QQmlEngine *, QJSEngine *) method");
+#endif
+
+        QQmlPrivate::qmlRegisterSingletonAndRevisions<Resolved, Extended, T>(
+                    uri, versionMajor, QQmlPrivate::StaticMetaObject<T>::staticMetaObject(),
+                    qmlTypeIds, extension);
+    }
+};
+
+template<class T, class Resolved, class Extended>
+struct QmlTypeAndRevisionsRegistration<T, Resolved, Extended, true, false, false, true> {
+    static void registerTypeAndRevisions(const char *uri, int versionMajor, QList<int> *qmlTypeIds,
+                                         const QMetaObject *extension)
+    {
+        // An uncreatable singleton makes little sense? OK, you can still use the enums.
+        QQmlPrivate::qmlRegisterSingletonAndRevisions<Resolved, Extended, T>(
                     uri, versionMajor, QQmlPrivate::StaticMetaObject<T>::staticMetaObject(),
                     qmlTypeIds, extension);
     }
 };
 
 template<class T, class Resolved>
-struct QmlTypeAndRevisionsRegistration<T, Resolved, void, false, true, false> {
+struct QmlTypeAndRevisionsRegistration<T, Resolved, void, false, true, false, false> {
     static void registerTypeAndRevisions(const char *uri, int versionMajor, QList<int> *qmlTypeIds,
                                          const QMetaObject *)
     {
@@ -858,36 +921,28 @@ struct QmlTypeAndRevisionsRegistration<T, Resolved, void, false, true, false> {
     }
 };
 
-template<typename T = void, typename... Args>
-void qmlRegisterTypesAndRevisions(const char *uri, int versionMajor,
-                                  QList<int> *qmlTypeIds = nullptr);
-
-template<typename T, typename... Args>
+template<typename... T>
 void qmlRegisterTypesAndRevisions(const char *uri, int versionMajor, QList<int> *qmlTypeIds)
 {
-    QmlTypeAndRevisionsRegistration<
+    (QmlTypeAndRevisionsRegistration<
             T, typename QQmlPrivate::QmlResolved<T>::Type,
             typename QQmlPrivate::QmlExtended<T>::Type,
             QQmlPrivate::QmlSingleton<T>::Value,
             QQmlPrivate::QmlInterface<T>::Value,
-            QQmlPrivate::QmlSequence<T>::Value>
+            QQmlPrivate::QmlSequence<T>::Value,
+            QQmlPrivate::QmlUncreatable<T>::Value || QQmlPrivate::QmlAnonymous<T>::Value>
             ::registerTypeAndRevisions(uri, versionMajor, qmlTypeIds,
-                                       QQmlPrivate::QmlExtendedNamespace<T>::metaObject());
-    qmlRegisterTypesAndRevisions<Args...>(uri, versionMajor, qmlTypeIds);
-}
-
-template<>
-inline void qmlRegisterTypesAndRevisions<>(const char *, int, QList<int> *)
-{
+                                       QQmlPrivate::QmlExtendedNamespace<T>::metaObject()), ...);
 }
 
 inline void qmlRegisterNamespaceAndRevisions(const QMetaObject *metaObject,
                                              const char *uri, int versionMajor,
-                                             QList<int> *qmlTypeIds = nullptr,
-                                             const QMetaObject *classInfoMetaObject = nullptr)
+                                             QList<int> *qmlTypeIds,
+                                             const QMetaObject *classInfoMetaObject,
+                                             const QMetaObject *extensionMetaObject)
 {
     QQmlPrivate::RegisterTypeAndRevisions type = {
-        0,
+        3,
         QMetaType(),
         QMetaType(),
         0,
@@ -909,15 +964,27 @@ inline void qmlRegisterNamespaceAndRevisions(const QMetaObject *metaObject,
         -1,
 
         nullptr,
-        nullptr,
+        extensionMetaObject,
 
         &qmlCreateCustomParser<void>,
-        qmlTypeIds
+        qmlTypeIds,
+        -1,
+        false,
+        QMetaSequence()
     };
 
     qmlregister(QQmlPrivate::TypeAndRevisionsRegistration, &type);
 }
 
+inline void qmlRegisterNamespaceAndRevisions(const QMetaObject *metaObject,
+                                             const char *uri, int versionMajor,
+                                             QList<int> *qmlTypeIds = nullptr,
+                                             const QMetaObject *classInfoMetaObject = nullptr)
+{
+    qmlRegisterNamespaceAndRevisions(metaObject, uri, versionMajor, qmlTypeIds,
+                                     classInfoMetaObject, nullptr);
+}
+
 int Q_QML_EXPORT qmlTypeId(const char *uri, int versionMajor, int versionMinor, const char *qmlName);
 
 QT_END_NAMESPACE
