@@ -1,41 +1,5 @@
-/****************************************************************************
-**
-** Copyright (C) 2016 The Qt Company Ltd.
-** Contact: https://www.qt.io/licensing/
-**
-** This file is part of the QtQml module of the Qt Toolkit.
-**
-** $QT_BEGIN_LICENSE:LGPL$
-** Commercial License Usage
-** Licensees holding valid commercial Qt licenses may use this file in
-** accordance with the commercial license agreement provided with the
-** Software or, alternatively, in accordance with the terms contained in
-** a written agreement between you and The Qt Company. For licensing terms
-** and conditions see https://www.qt.io/terms-conditions. For further
-** information use the contact form at https://www.qt.io/contact-us.
-**
-** GNU Lesser General Public License Usage
-** Alternatively, this file may be used under the terms of the GNU Lesser
-** General Public License version 3 as published by the Free Software
-** Foundation and appearing in the file LICENSE.LGPL3 included in the
-** packaging of this file. Please review the following information to
-** ensure the GNU Lesser General Public License version 3 requirements
-** will be met: https://www.gnu.org/licenses/lgpl-3.0.html.
-**
-** GNU General Public License Usage
-** Alternatively, this file may be used under the terms of the GNU
-** General Public License version 2.0 or (at your option) the GNU General
-** Public license version 3 or any later version approved by the KDE Free
-** Qt Foundation. The licenses are as published by the Free Software
-** Foundation and appearing in the file LICENSE.GPL2 and LICENSE.GPL3
-** included in the packaging of this file. Please review the following
-** information to ensure the GNU General Public License requirements will
-** be met: https://www.gnu.org/licenses/gpl-2.0.html and
-** https://www.gnu.org/licenses/gpl-3.0.html.
-**
-** $QT_END_LICENSE$
-**
-****************************************************************************/
+// Copyright (C) 2016 The Qt Company Ltd.
+// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR LGPL-3.0-only OR GPL-2.0-only OR GPL-3.0-only
 
 #ifndef QQMLPRIVATE_H
 #define QQMLPRIVATE_H
@@ -141,13 +105,16 @@ namespace QQmlPrivate
             // the size that was allocated.
             ::operator delete (ptr);
         }
+#ifdef Q_CC_MSVC
         static void operator delete(void *, void *) {
             // Deliberately empty placement delete operator.
             // Silences MSVC warning C4291: no matching operator delete found
+            // On MinGW it causes -Wmismatched-new-delete, though.
         }
+#endif
     };
 
-    enum class ConstructionMode
+    enum class SingletonConstructionMode
     {
         None,
         Constructor,
@@ -172,33 +139,45 @@ namespace QQmlPrivate
     };
 
     template<typename T, typename WrapperT>
-    constexpr ConstructionMode constructionMode()
+    constexpr SingletonConstructionMode singletonConstructionMode()
     {
         if constexpr (!std::is_base_of<QObject, T>::value)
-            return ConstructionMode::None;
+            return SingletonConstructionMode::None;
         if constexpr (!std::is_same_v<T, WrapperT> && HasSingletonFactory<T, WrapperT>::value)
-            return ConstructionMode::FactoryWrapper;
+            return SingletonConstructionMode::FactoryWrapper;
         if constexpr (std::is_default_constructible<T>::value)
-            return ConstructionMode::Constructor;
+            return SingletonConstructionMode::Constructor;
         if constexpr (HasSingletonFactory<T>::value)
-            return ConstructionMode::Factory;
+            return SingletonConstructionMode::Factory;
 
-        return ConstructionMode::None;
+        return SingletonConstructionMode::None;
     }
 
+    template<typename>
+    struct QmlMarkerFunction;
+
+    template<typename Ret, typename Class>
+    struct QmlMarkerFunction<Ret (Class::*)()>
+    {
+        using ClassType = Class;
+    };
+
+    template<typename T, typename Marker>
+    using QmlTypeHasMarker = std::is_same<T, typename QmlMarkerFunction<Marker>::ClassType>;
+
     template<typename T>
     void createInto(void *memory, void *) { new (memory) QQmlElement<T>; }
 
-    template<typename T, typename WrapperT, ConstructionMode Mode>
+    template<typename T, typename WrapperT, SingletonConstructionMode Mode>
     QObject *createSingletonInstance(QQmlEngine *q, QJSEngine *j)
     {
         Q_UNUSED(q);
         Q_UNUSED(j);
-        if constexpr (Mode == ConstructionMode::Constructor)
+        if constexpr (Mode == SingletonConstructionMode::Constructor)
             return new T;
-        else if constexpr (Mode == ConstructionMode::Factory)
+        else if constexpr (Mode == SingletonConstructionMode::Factory)
             return T::create(q, j);
-        else if constexpr (Mode == ConstructionMode::FactoryWrapper)
+        else if constexpr (Mode == SingletonConstructionMode::FactoryWrapper)
             return WrapperT::create(q, j);
         else
             return nullptr;
@@ -212,39 +191,43 @@ namespace QQmlPrivate
     using CreateParentFunction = QObject *(*)(QObject *);
     using CreateValueTypeFunction = QVariant (*)(const QJSValue &);
 
-    template<typename T, typename WrapperT = T, ConstructionMode Mode = constructionMode<T, WrapperT>()>
+    template<typename T, typename WrapperT = T,
+             SingletonConstructionMode Mode = singletonConstructionMode<T, WrapperT>()>
     struct Constructors;
 
     template<typename T, typename WrapperT>
-    struct Constructors<T, WrapperT, ConstructionMode::Constructor>
+    struct Constructors<T, WrapperT, SingletonConstructionMode::Constructor>
     {
         static constexpr CreateIntoFunction createInto
                 = QQmlPrivate::createInto<T>;
         static constexpr CreateSingletonFunction createSingletonInstance
-                = QQmlPrivate::createSingletonInstance<T, WrapperT, ConstructionMode::Constructor>;
+                = QQmlPrivate::createSingletonInstance<
+                    T, WrapperT, SingletonConstructionMode::Constructor>;
     };
 
     template<typename T, typename WrapperT>
-    struct Constructors<T, WrapperT, ConstructionMode::None>
+    struct Constructors<T, WrapperT, SingletonConstructionMode::None>
     {
         static constexpr CreateIntoFunction createInto = nullptr;
         static constexpr CreateSingletonFunction createSingletonInstance = nullptr;
     };
 
     template<typename T, typename WrapperT>
-    struct Constructors<T, WrapperT, ConstructionMode::Factory>
+    struct Constructors<T, WrapperT, SingletonConstructionMode::Factory>
     {
         static constexpr CreateIntoFunction createInto = nullptr;
         static constexpr CreateSingletonFunction createSingletonInstance
-                = QQmlPrivate::createSingletonInstance<T, WrapperT, ConstructionMode::Factory>;
+                = QQmlPrivate::createSingletonInstance<
+                    T, WrapperT, SingletonConstructionMode::Factory>;
     };
 
     template<typename T, typename WrapperT>
-    struct Constructors<T, WrapperT, ConstructionMode::FactoryWrapper>
+    struct Constructors<T, WrapperT, SingletonConstructionMode::FactoryWrapper>
     {
         static constexpr CreateIntoFunction createInto = nullptr;
         static constexpr CreateSingletonFunction createSingletonInstance
-                = QQmlPrivate::createSingletonInstance<T, WrapperT, ConstructionMode::FactoryWrapper>;
+                = QQmlPrivate::createSingletonInstance<
+                    T, WrapperT, SingletonConstructionMode::FactoryWrapper>;
     };
 
     template<typename T,
@@ -382,11 +365,13 @@ namespace QQmlPrivate
         struct Properties<Parent, void>
         {
             using Func = QQmlAttachedPropertiesFunc<QObject>;
-            static const QMetaObject *staticMetaObject() { return nullptr; };
-            static Func attachedPropertiesFunc() { return nullptr; };
+            static const QMetaObject *staticMetaObject() { return nullptr; }
+            static Func attachedPropertiesFunc() { return nullptr; }
         };
 
-        using Type = typename OverridableAttachedType<T, typename T::QmlAttachedType>::Type;
+        using Type = typename std::conditional<
+                QmlTypeHasMarker<T, decltype(&T::qt_qmlMarker_attached)>::value,
+                typename OverridableAttachedType<T, typename T::QmlAttachedType>::Type, void>::type;
         using Func = typename Properties<T, Type>::Func;
 
         static const QMetaObject *staticMetaObject()
@@ -446,7 +431,18 @@ namespace QQmlPrivate
     enum AutoParentResult { Parented, IncompatibleObject, IncompatibleParent };
     typedef AutoParentResult (*AutoParentFunction)(QObject *object, QObject *parent);
 
+    enum class ValueTypeCreationMethod { None, Construct, Structured };
+
     struct RegisterType {
+        enum StructVersion: int {
+            Base = 0,
+            FinalizerCast = 1,
+            CreationMethod = 2,
+            CurrentVersion = CreationMethod,
+        };
+
+        bool has(StructVersion v) const { return structVersion >= int(v); }
+
         int structVersion;
 
         QMetaType typeId;
@@ -457,6 +453,7 @@ namespace QQmlPrivate
         void *userdata;
         QString noCreationReason;
 
+        // ### Qt7: Get rid of this. It can be covered by creationMethod below.
         QVariant (*createValueType)(const QJSValue &);
 
         const char *uri;
@@ -478,6 +475,8 @@ namespace QQmlPrivate
 
         QTypeRevision revision;
         int finalizerCast;
+
+        ValueTypeCreationMethod creationMethod;
         // If this is extended ensure "version" is bumped!!!
     };
 
@@ -513,6 +512,7 @@ namespace QQmlPrivate
         int finalizerCast;
 
         bool forceAnonymous;
+        QMetaSequence listMetaSequence;
     };
 
     struct RegisterInterface {
@@ -590,7 +590,11 @@ namespace QQmlPrivate
         int structVersion;
         const char *uri;
         QTypeRevision version;
+
+        // ### Qt7: Remove typeName. It's ignored because the only valid name is "list",
+        //          and that's automatic.
         const char *typeName;
+
         QMetaType typeId;
         QMetaSequence metaSequence;
         QTypeRevision revision;
@@ -614,8 +618,12 @@ namespace QQmlPrivate
         QQmlContextData *qmlContext;
         QObject *qmlScopeObject;
         QJSEngine *engine;
-        QV4::ExecutableCompilationUnit *compilationUnit;
+        union {
+            QV4::ExecutableCompilationUnit *compilationUnit;
+            qintptr extraData;
+        };
 
+        QObject *thisObject() const;
         QQmlEngine *qmlEngine() const;
 
         QJSValue jsMetaType(int index) const;
@@ -625,10 +633,22 @@ namespace QQmlPrivate
         // Run QQmlPropertyCapture::captureProperty() without retrieving the value.
         bool captureLookup(uint index, QObject *object) const;
         bool captureQmlContextPropertyLookup(uint index) const;
+        void captureTranslation() const;
+        QString translationContext() const;
         QMetaType lookupResultMetaType(uint index) const;
         void storeNameSloppy(uint nameIndex, void *value, QMetaType type) const;
         QJSValue javaScriptGlobalProperty(uint nameIndex) const;
 
+        const QLoggingCategory *resolveLoggingCategory(QObject *wrapper, bool *ok) const;
+
+        void writeToConsole(
+                QtMsgType type, const QString &message,
+                const QLoggingCategory *loggingCategory) const;
+
+        QVariant constructValueType(
+                QMetaType resultMetaType, const QMetaObject *resultMetaObject,
+                int ctorIndex, void *ctorArg) const;
+
         // All of these lookup functions should be used as follows:
         //
         // while (!fooBarLookup(...)) {
@@ -683,7 +703,10 @@ namespace QQmlPrivate
         bool getValueLookup(uint index, void *value, void *target) const;
         void initGetValueLookup(uint index, const QMetaObject *metaObject, QMetaType type) const;
 
+        bool getEnumLookup(uint index, void *target) const;
+#if QT_QML_REMOVED_SINCE(6, 6)
         bool getEnumLookup(uint index, int *target) const;
+#endif
         void initGetEnumLookup(uint index, const QMetaObject *metaObject,
                                const char *enumerator, const char *enumValue) const;
 
@@ -695,12 +718,17 @@ namespace QQmlPrivate
     };
 
     struct AOTCompiledFunction {
-        int index;
+        qintptr extraData;
         QMetaType returnType;
         QList<QMetaType> argumentTypes;
         void (*functionPtr)(const AOTCompiledContext *context, void *resultPtr, void **arguments);
     };
 
+#if QT_DEPRECATED_SINCE(6, 6)
+    QT_DEPRECATED_VERSION_X(6, 6, "Use AOTCompiledFunction instead")
+    typedef AOTCompiledFunction TypedFunction;
+#endif
+
     struct CachedQmlUnit {
         const QV4::CompiledData::Unit *qmlData;
         const AOTCompiledFunction *aotCompiledFunctions;
@@ -778,26 +806,18 @@ namespace QQmlPrivate
         const int index = indexOfOwnClassInfo(metaObject, key);
         return (index == -1) ? defaultValue
                              : QTypeRevision::fromEncodedVersion(
-                                   QByteArray(metaObject->classInfo(index).value()).toInt());
+                                   QLatin1StringView(metaObject->classInfo(index).value()).toInt());
     }
 
-    inline QList<QTypeRevision> revisionClassInfos(const QMetaObject *metaObject, const char *key)
-    {
-        QList<QTypeRevision> revisions;
-        for (int index = indexOfOwnClassInfo(metaObject, key); index != -1;
-             index = indexOfOwnClassInfo(metaObject, key, index - 1)) {
-            revisions.push_back(QTypeRevision::fromEncodedVersion(
-                          QByteArray(metaObject->classInfo(index).value()).toInt()));
-        }
-        return revisions;
-    }
+    Q_QML_EXPORT QList<QTypeRevision> revisionClassInfos(const QMetaObject *metaObject, const char *key);
 
     inline bool boolClassInfo(const QMetaObject *metaObject, const char *key,
                               bool defaultValue = false)
     {
         const int index = indexOfOwnClassInfo(metaObject, key);
-        return (index == -1) ? defaultValue
-                             : (QByteArray(metaObject->classInfo(index).value()) == "true");
+        if (index == -1)
+            return defaultValue;
+        return qstrcmp(metaObject->classInfo(index).value(), "true") == 0;
     }
 
     inline const char *classElementName(const QMetaObject *metaObject)
@@ -832,7 +852,9 @@ namespace QQmlPrivate
     template<class T>
     struct QmlExtended<T, std::void_t<typename T::QmlExtendedType>>
     {
-        using Type = typename T::QmlExtendedType;
+        using Type = typename std::conditional<
+                QmlTypeHasMarker<T, decltype(&T::qt_qmlMarker_extended)>::value,
+                typename T::QmlExtendedType, void>::type;
     };
 
     template<class T, class = std::void_t<>>
@@ -844,7 +866,13 @@ namespace QQmlPrivate
     template<class T>
     struct QmlExtendedNamespace<T, std::void_t<decltype(T::qmlExtendedNamespace())>>
     {
-        static constexpr const QMetaObject *metaObject() { return T::qmlExtendedNamespace(); }
+        static constexpr const QMetaObject *metaObject()
+        {
+            if constexpr (QmlTypeHasMarker<T, decltype(&T::qt_qmlMarker_extendedNamespace)>::value)
+                return T::qmlExtendedNamespace();
+            else
+                return nullptr;
+        }
     };
 
     template<class T, class = std::void_t<>>
@@ -856,9 +884,40 @@ namespace QQmlPrivate
     template<class T>
     struct QmlResolved<T, std::void_t<typename T::QmlForeignType>>
     {
-        using Type = typename T::QmlForeignType;
+        using Type = typename std::conditional<
+                QmlTypeHasMarker<T, decltype(&T::qt_qmlMarker_foreign)>::value,
+                typename T::QmlForeignType, T>::type;
     };
 
+    template<class T, class = std::void_t<>>
+    struct QmlUncreatable
+    {
+        static constexpr bool Value = false;
+    };
+
+    template<class T>
+    struct QmlUncreatable<T, std::void_t<typename T::QmlIsUncreatable>>
+    {
+        static constexpr bool Value =
+                QmlTypeHasMarker<T, decltype(&T::qt_qmlMarker_uncreatable)>::value
+                && bool(T::QmlIsUncreatable::yes);
+    };
+
+    template<class T, class = std::void_t<>>
+    struct QmlAnonymous
+    {
+        static constexpr bool Value = false;
+    };
+
+    template<class T>
+    struct QmlAnonymous<T, std::void_t<typename T::QmlIsAnonymous>>
+    {
+        static constexpr bool Value =
+                QmlTypeHasMarker<T, decltype(&T::qt_qmlMarker_anonymous)>::value
+                && bool(T::QmlIsAnonymous::yes);
+    };
+
+
     template<class T, class = std::void_t<>>
     struct QmlSingleton
     {
@@ -868,7 +927,9 @@ namespace QQmlPrivate
     template<class T>
     struct QmlSingleton<T, std::void_t<typename T::QmlIsSingleton>>
     {
-        static constexpr bool Value = bool(T::QmlIsSingleton::yes);
+        static constexpr bool Value =
+                QmlTypeHasMarker<T, decltype(&T::qt_qmlMarker_singleton)>::value
+                && bool(T::QmlIsSingleton::yes);
     };
 
     template<class T, class = std::void_t<>>
@@ -912,6 +973,16 @@ namespace QQmlPrivate
     template<class T>
     struct QmlMetaType
     {
+        static constexpr bool hasAcceptableCtors()
+        {
+            if constexpr (!std::is_default_constructible_v<T>)
+                return false;
+            else if constexpr (std::is_base_of_v<QObject, T>)
+                return true;
+            else
+                return std::is_copy_constructible_v<T>;
+        }
+
         static QMetaType self()
         {
             if constexpr (std::is_base_of_v<QObject, T>)
@@ -925,7 +996,15 @@ namespace QQmlPrivate
             if constexpr (std::is_base_of_v<QObject, T>)
                 return QMetaType::fromType<QQmlListProperty<T>>();
             else
-                return QMetaType();
+                return QMetaType::fromType<QList<T>>();
+        }
+
+        static QMetaSequence sequence()
+        {
+            if constexpr (std::is_base_of_v<QObject, T>)
+                return QMetaSequence();
+            else
+                return QMetaSequence::fromContainer<QList<T>>();
         }
     };
 
@@ -934,6 +1013,7 @@ namespace QQmlPrivate
                                           const QMetaObject *classInfoMetaObject,
                                           QVector<int> *qmlTypeIds, const QMetaObject *extension)
     {
+        static_assert(std::is_base_of_v<QObject, T>);
         RegisterSingletonTypeAndRevisions api = {
             0,
 
@@ -963,7 +1043,7 @@ namespace QQmlPrivate
                                      bool forceAnonymous = false)
     {
         RegisterTypeAndRevisions type = {
-            2,
+            3,
             QmlMetaType<T>::self(),
             QmlMetaType<T>::list(),
             int(sizeof(T)),
@@ -991,7 +1071,8 @@ namespace QQmlPrivate
             qmlTypeIds,
             StaticCastSelector<T, QQmlFinalizerHook>::cast(),
 
-            forceAnonymous
+            forceAnonymous,
+            QmlMetaType<T>::sequence(),
         };
 
         // Initialize the extension so that we can find it by name or ID.
@@ -1047,6 +1128,16 @@ namespace QQmlPrivate
         };
     }
 
+    Q_QML_EXPORT QObject *qmlExtendedObject(QObject *, int);
+
+    enum QmlRegistrationWarning {
+        UnconstructibleType,
+        UnconstructibleSingleton,
+        NonQObjectWithAtached,
+    };
+
+    Q_QML_EXPORT void qmlRegistrationWarning(QmlRegistrationWarning warning, QMetaType type);
+
 } // namespace QQmlPrivate
 
 QT_END_NAMESPACE
