@@ -33,18 +33,29 @@
 
 QT_BEGIN_NAMESPACE
 
+namespace QQmlJS::Dom {
+class QQmlDomAstCreatorWithQQmlJSScope;
+}
+
 struct QQmlJSResourceFileMapper;
 class Q_QMLCOMPILER_PRIVATE_EXPORT QQmlJSImportVisitor : public QQmlJS::AST::Visitor
 {
 public:
+    QQmlJSImportVisitor();
     QQmlJSImportVisitor(const QQmlJSScope::Ptr &target,
                         QQmlJSImporter *importer, QQmlJSLogger *logger,
                         const QString &implicitImportDirectory,
                         const QStringList &qmldirFiles = QStringList());
     ~QQmlJSImportVisitor();
 
+    using QQmlJS::AST::Visitor::endVisit;
+    using QQmlJS::AST::Visitor::postVisit;
+    using QQmlJS::AST::Visitor::preVisit;
+    using QQmlJS::AST::Visitor::visit;
+
     QQmlJSScope::Ptr result() const { return m_exportedRootScope; }
 
+    const QQmlJSLogger *logger() const { return m_logger; }
     QQmlJSLogger *logger() { return m_logger; }
 
     QQmlJSImporter::ImportedTypes imports() const { return m_rootScopeImports; }
@@ -63,7 +74,9 @@ public:
     static QString implicitImportDirectory(
             const QString &localFile, QQmlJSResourceFileMapper *mapper);
 
-    QQmlJSImporter *importer() { return m_importer; } // ### should this be restricted?
+    // ### should this be restricted?
+    QQmlJSImporter *importer() { return m_importer; }
+    const QQmlJSImporter *importer() const { return m_importer; }
 
     struct UnfinishedBinding
     {
@@ -246,6 +259,7 @@ protected:
     void processPropertyBindings();
     void checkRequiredProperties();
     void processPropertyTypes();
+    void processMethodTypes();
     void processPropertyBindingObjects();
     void flushPendingSignalParameters();
 
@@ -271,6 +285,13 @@ protected:
         QQmlJS::SourceLocation location;
     };
 
+    struct PendingMethodType
+    {
+        QQmlJSScope::Ptr scope;
+        QString methodName;
+        QQmlJS::SourceLocation location;
+    };
+
     struct PendingPropertyObjectBinding
     {
         QQmlJSScope::Ptr scope;
@@ -308,6 +329,7 @@ protected:
 
     QHash<QQmlJSScope::Ptr, QVector<QQmlJSScope::Ptr>> m_pendingDefaultProperties;
     QVector<PendingPropertyType> m_pendingPropertyTypes;
+    QVector<PendingMethodType> m_pendingMethodTypes;
     QVector<PendingPropertyObjectBinding> m_pendingPropertyObjectBindings;
     QVector<RequiredProperty> m_requiredProperties;
     QVector<QQmlJSScope::Ptr> m_objectBindingScopes;
@@ -335,12 +357,15 @@ private:
     void populateCurrentScope(QQmlJSScope::ScopeType type, const QString &name,
                               const QQmlJS::SourceLocation &location);
     void enterRootScope(QQmlJSScope::ScopeType type, const QString &name,
-                        const QQmlJS::SourceLocation &location);
+                           const QQmlJS::SourceLocation &location);
 
     void importFromHost(const QString &path, const QString &prefix,
                         const QQmlJS::SourceLocation &location);
     void importFromQrc(const QString &path, const QString &prefix,
                        const QQmlJS::SourceLocation &location);
+
+public:
+    friend class QQmlJS::Dom::QQmlDomAstCreatorWithQQmlJSScope;
 };
 
 QT_END_NAMESPACE
