@@ -8,9 +8,9 @@
 
 #define QT_FEATURE_quickcontrols2_universal 1
 
-#define QT_FEATURE_quickcontrols2_macos 1
+#define QT_FEATURE_quickcontrols2_macos -1
 
-#define QT_FEATURE_quickcontrols2_ios 1
+#define QT_FEATURE_quickcontrols2_ios -1
 
-#define QT_FEATURE_quickcontrols2_windows 1
+#define QT_FEATURE_quickcontrols2_windows -1
 
